#chartconfig
"""Allows charts to be overridden by configuration files.

This feature is an addition in November 2005.  I've tried to implement
it with minimum changes to the code base.

Here's the kind of file you can create:

    ---------------------------diagra.ini--------------------------------

    [DEFAULT]
    #this will apply to all chart types. So, it's a great place
    #to set database connection parameters.
    dataSource.name = 'mydata_uat'
    dataSource.user = 'gordon'
    dataSource.password = 'drongo'



    [module]
    #any in here will apply to modules with corresponding names. This is
    #the usual way to do it, as end users will not see the class names
    #very often.

    #specifics in module override stuff in DEFAULT


Lots of doctests follow.  Start by making a drawing

>>> from reportlab.graphics.shapes import Drawing
>>> from reportlab.graphics.charts.barcharts import HorizontalBarChart
>>> d = Drawing()
>>> chart = HorizontalBarChart()
>>> d.add(chart, name='chart')
>>> d.height
200
>>> d.chart.valueAxis.labels.fontName
'Times-Roman'

>>> d.__class__.__module__
'reportlab.graphics.shapes'
>>> #hack it so I can test better
>>> d.__class__.__module__ = 'mychart'
>>> d.__class__.__module__
'mychart'

>>> configData = '''
... [DEFAULT]
... dataSource.name = 'mydata_uat'
... dataSource.user = 'gordon'
... dataSource.password = 'drongo'
...
... [mychart]
... chart.valueAxis.labels.fontName='Helvetica'
... chart.valueAxis.labels.fontSize=12
... someColor = red
... height = 250
... '''
>>> #parse it by hand
>>> from reportlab.lib.utils import getStringIO
>>> config = SafeConfigParser()
>>> config.optionxform = str
>>> buf = getStringIO(configData)
>>> config.readfp(buf)
>>> del buf
>>> config.has_section('mychart')
True
>>> from pprint import pprint as pp
>>> pp(sorted(config.items('DEFAULT')))
[('dataSource.name', "'mydata_uat'"),
 ('dataSource.password', "'drongo'"),
 ('dataSource.user', "'gordon'")]
>>> pp(sorted(config.items('mychart')))
[('chart.valueAxis.labels.fontName', "'Helvetica'"),
 ('chart.valueAxis.labels.fontSize', '12'),
 ('dataSource.name', "'mydata_uat'"),
 ('dataSource.password', "'drongo'"),
 ('dataSource.user', "'gordon'"),
 ('height', '250'),
 ('someColor', 'red')]
>>> applied = applyToDrawing(config, d)
>>> applied
3
>>> d.height
250
"""
import sys, os
try:
    from ConfigParser import SafeConfigParser
except ImportError:
    from configparser import SafeConfigParser

def findConfigFilesUpwards(fileName, startDir=None, maxLevels=None):
    """Find and return list of full paths to config file, searching up.

    If startDir not provided, take current directory.
    if maxLevels not set, it tracks up to the top of the file system.

    If not found, return empty list"""
    if startDir is None:
        startDir = os.getcwd()

    found = []
    dirName = startDir
    levels = 1
    while 1:
        searchFor = dirName + os.sep + fileName
        if os.path.isfile(searchFor):
            found.append(searchFor)
        newDirName = os.path.dirname(dirName)
        if newDirName == dirName:
            break
        else:
            dirName = newDirName
            levels += 1
            if (maxLevels is not None) and (levels > maxLevels):
                break

    return found

def getConfig(fileName, startDir=None, maxLevels=None, verbose=False):
    """Returns a parsed configParser object using all config files
    found above the search path.

    """
    confFileNames = findConfigFilesUpwards(fileName, startDir, maxLevels)
    cp = SafeConfigParser()
    cp.optionxform = str   #make it case sensitive
    if confFileNames:
        #found it/them
        #the one closest to start has highest priority
        #so read the highest first
        confFileNames.reverse()
        for confFileName in confFileNames:
            if verbose:
                print('found config %s' % confFileName)
            cp.read(confFileName)
    return cp

def applyToDrawing(config, drawing, verbose=False):
    """Applies all relevant info to the drawing.

    If verbose, reports 'misses' to alert you to any
    config entries not used.
    """
    if config is None:
        return

    if verbose:
        print('Applying config options:')

    items =dict([(k,v) for k,v in config.items('DEFAULT')])
    moduleName = drawing.__class__.__module__
    if moduleName=='__main__':
        moduleName = os.path.splitext(os.path.basename(sys.argv[0]))[0]
    if verbose: print('module name=%r' % moduleName)
    if config.has_section(moduleName):
        #this is intended for us.  The configparser puts all
        #the default in here anyway
        items.update(dict([(k,v) for k,v in config.items(moduleName)]))
    if verbose: print('config items=%r' % items)

    #we really want the namespace of the chart module, but if
    try:
        ns = sys.modules[moduleName].__dict__.copy()
        if verbose: print('Using namespace from', moduleName)
    except:
        ns = {}
    #that fails we want one allowing colour imports at the least.
    #presumably Robin could apply here whatever trickery he did
    #in guiedit to type-convert

    from reportlab.graphics.shapes import STATE_DEFAULTS
    SDK = list(STATE_DEFAULTS.keys())
    ns.update(dict(drawing=drawing, STATE_DEFAULTS=STATE_DEFAULTS))
    exec("from reportlab.lib.colors import *", ns)

    setOK = 0
    for fullPath, strValue in items.items():
        if fullPath in SDK:
            command = 'STATE_DEFAULTS[%r] = %s' % (fullPath,strValue)
        else:
            command = 'drawing.%s = %s' % (fullPath, strValue)
        try:
            if verbose and 'password' not in command:
                sys.stdout.write('  trying %s'%command)
            exec(command, ns)
            setOK += 1
            if verbose: print('...OK')
        except Exception as e:
            if verbose: print('...failed: %s' % e)
    if verbose: print()

    return setOK

if __name__=='__main__':
    import doctest, chartconfig
    doctest.testmod(chartconfig)



__code__='Gat=(D/[on&>$RVN[.F,+f2#hiptM=Q0KU,WN)tuoXM8!c8FS_<unCoa0S?2I(!2H-5S+Y$<"Y2=7qo&Ao2UU+r)QKH.Lj.H!\\MIM.#K73aFtt<W#2bDo)e<(DMB"Dh6qg"%E?fJ7IN,a^6/0JFmUY3e2:%/\\j/mo7.4!Mh#F@cN/ItGTQ>Ui"Fs:^h(;SJ1H-l"\'u30"Ro@*.tF5mIL^T;n/C4$JCfY>_Q,#)4o+\\!r\'B*JN93C6=FkQZcG=K6`9Y0HpsO_[`9m\\Ua\\kX)i<B_E\\&a:9\'S_aK/FqGhlY!YOPJRns:KfhjE)_,qD!J-16@\'WYSLI=?_VqIS]"3j\\F1DDn$+To1$/0tP2&V&M(;GH=e2\'fL=Z^TPJGaE#<H)N1"c1Zo\\KpO2ZP8&H7Bl=KBn*J`A8SZtN91P&*N;E.Q/RS-0I`fg/ZMrKhc*)!j]od(/hK&+:U4iNUkZRAM]L9>E@Yn4/K;nS$7_/OXho#FHOmd>LRgb#SSG07FIX$cZ]uj#jbG`WcP`gO[j>>pb@4%F[\\9&NrDGPW"[O&eAIt4tq_6=B,lDo45/&p`;Q\\[TA>uZGgFTj,9!X6]#JmD4g)%l@"U>[_dN8u(At;EcRERI`rG]s9n=L6H\\8XeoM\'JE(\'&?fTi^g]7p<:6VXn2!CSV)Os:YEi8h"%1Um("E_h_uU9^&9W>ToQJ/c\'0_+H)q9+b4<H6E(ip(:7:Ad[b/ih1:aQAHn_U:gmP5Sm9V`!&H60e?`J^cl7@"A,[1.!@GNk$:)^,LhBOHIDjYlP9r^R\\h/lTm]oq1p3]AuIg\'nGSQ0IU_F.UBuG?=7]n\'UF<#ES;&="QQob9rN=9smqLag+6[^N[q(F;\'^\\I7h`Cr2PogiE!E6ACib51e,)4g5FFN]&MhGhm7_\'>H$H<m@r\\5]5)"iB&/\'KgL6UCX8m^U+*"t;gT/\'%(D_3;,GS^rke;kZ`"k824#;AmXh*?Xf/3;lS&[$fHfH\'D>)]Vr&,27lou9E;g4mK?rL#iMSOSP_9rIAaS!`RpjdG-_3V?6Zaalmp`kia3KhgEF^-AG3[ccuM(\'J\'/OK_V5Vm0.#1Y-SOfS)I-pJ5nF.I^=.BnWgU+3O<L,I8*An-]eo=pne*[/T7g8n1;s)ZKjdJ"in~>'



from io import StringIO

import unittest
import unittest2

from unittest2.test.support import resultFactory


class TestUnittest(unittest2.TestCase):
    
    def assertIsSubclass(self, actual, klass):
        self.assertTrue(issubclass(actual, klass), "Not a subclass.")
    
    def testInheritance(self):
        self.assertIsSubclass(unittest2.TestCase, unittest.TestCase)
        self.assertIsSubclass(unittest2.TestResult, unittest.TestResult)
        self.assertIsSubclass(unittest2.TestSuite, unittest.TestSuite)
        self.assertIsSubclass(unittest2.TextTestRunner, unittest.TextTestRunner)
        self.assertIsSubclass(unittest2.TestLoader, unittest.TestLoader)
        self.assertIsSubclass(unittest2.TextTestResult, unittest.TestResult)
    
    def test_new_runner_old_case(self):
        runner = unittest2.TextTestRunner(resultclass=resultFactory,
                                          stream=StringIO())
        class Test(unittest.TestCase):
            def testOne(self):
                pass
        suite = unittest2.TestSuite((Test('testOne'),))
        result = runner.run(suite)
        self.assertEqual(result.testsRun, 1)
        self.assertEqual(len(result.errors), 0)
    
    def test_old_runner_new_case(self):
        runner = unittest.TextTestRunner(stream=StringIO())
        class Test(unittest2.TestCase):
            def testOne(self):
                self.assertDictEqual({}, {})
                
        suite = unittest.TestSuite((Test('testOne'),))
        result = runner.run(suite)
        self.assertEqual(result.testsRun, 1)
        self.assertEqual(len(result.errors), 0)
        
        
if __name__ == '__main__':
    unittest2.main()


__code__='Gas1]gMYb8&=k*r=]X<G;;ANmPD7>YN=(o-F$ie=70>sb%*5F\\-#pc.+DB\'SAk\\?2Eb3FIoAfn"l&YHqqAF\\J?hC<RZl#i8@;0qig7:6:B!c4NL:PW1RQo3DrZ[!H_$,(8!YF9AYWXbW\'d*Z"!AC=bK/<RBR",r[%Bp>P!8%9V"De.2!F,^(#k0D\'3PZEgH,)>r"i+]bNr^m8IKK9ni$H1b!\'^_$%!VsF$c&g]Epf47>`X?G%)3Fa"`=c[SMQ?Ca;Q%d+C>\'K)_Wdf8aJ@nQ^k\'K3F%26k=-UQK)#JmW^*^dQ:4VQ/la$+FhAIIRf%6VZ\'6B7@:7B4XkE%q^aaP!^nH::DbSLXLGV[/Tq&DPBMn/aAX;&#6NbblQ)ihT*XNCr:K570N-@)Let9)TTCj9hIb$63Kl*f;X];Wfbnd]dKsC&JKpBbnWt98ZNjETSfSoh5eY%Tpl)7-J[IV)GleHP^mWUH%FD"h6M%6;^45!:OFOGfY85h,lj#@TEd2"<r>8p>+&jaG,@XfplY@r7=35J2)lQsl(4n5<-\'@ingP\',&$ZJhr(%*`BU\\E:]V-ss<QatutAi,@?8KBFI@b\\Z0s_Ec,/6p"C:&njA[Re,:]3rQeIh;@+6<4NnBRnf\'(*Lea<KA>_[q9d.4@1q=OZK3nP\'k%8;HLP8a-DVcBJd2%Q?D>ppeJ7oErk7*ir(luibHJZ^i1<mcB;Um2<g\'ERZ0U*d^G_GR%.s^!Wc1,9dn=0;;dW:[Dgfum%T1mC6j5Z$-63Fu)nK\\Ngg7Ncj2&n:HOoOna.(EHp\'JlVXJLgQL_B"Qd`]B4a*_no_at8KrUD[j[Nq6s+E%EF1>!&rmAbfG4m;(rWh7iAE&/a2618g-r3"KgBQ*]T,a!_.cC<?]ZHN"\'7G]*-@%ou9@(uu2W,0jC"nT^5<s!Kb9YY1"]b9E/nHOM47^g_i/\'OXkU\'Vd8b/A7rBtiIA#+BEa$X3\'hm>+)&hn.$"h(\'dqKaFk:KpGOeIMFo<7-ZUd"jce8LnI@NMfFYYW9#uh`,6al\'[*QdqEMA%N?=2Roco6)hf+G36@D]^MY\\dZ9?+99)ufuTVe!+~>'



"""
This is a collection of utilities used by the ``svglib`` code module.
"""
from reportlab import xrange
import re
from math import acos, ceil, copysign, cos, degrees, fabs, hypot, radians, sin, sqrt

from reportlab.graphics.shapes import mmult, rotate, translate, transformPoint


def split_floats(op, min_num, value):
    """Split `value`, a list of numbers as a string, to a list of float numbers.

    Also optionally insert a `l` or `L` operation depending on the operation
    and the length of values.
    Example: with op='m' and value='10,20 30,40,' the returned value will be
             ['m', [10.0, 20.0], 'l', [30.0, 40.0]]
    """
    floats = [float(seq) for seq in re.findall('(-?\d*\.?\d*(?:e[+-]\d+)?)', value) if seq]
    res = []
    for i in range(0, len(floats), min_num):
        if i > 0 and op in {'m', 'M'}:
            op = 'l' if op == 'm' else 'L'
        res.extend([op, floats[i:i + min_num]])
    return res


def normalise_svg_path(attr):
    """Normalise SVG path.

    This basically introduces operator codes for multi-argument
    parameters. Also, it fixes sequences of consecutive M or m
    operators to MLLL... and mlll... operators. It adds an empty
    list as argument for Z and z only in order to make the resul-
    ting list easier to iterate over.

    E.g. "M 10 20, M 20 20, L 30 40, 40 40, Z"
      -> ['M', [10, 20], 'L', [20, 20], 'L', [30, 40], 'L', [40, 40], 'Z', []]
    """

    # operator codes mapped to the minimum number of expected arguments
    ops = {
        'A': 7, 'a': 7,
        'Q': 4, 'q': 4, 'T': 2, 't': 2, 'S': 4, 's': 4,
        'M': 2, 'L': 2, 'm': 2, 'l': 2, 'H': 1, 'V': 1,
        'h': 1, 'v': 1, 'C': 6, 'c': 6, 'Z': 0, 'z': 0,
    }
    op_keys = ops.keys()

    # do some preprocessing
    result = []
    groups = re.split('([achlmqstvz])', attr.strip(), flags=re.I)
    op = None
    for item in groups:
        if item.strip() == '':
            continue
        if item in op_keys:
            # fix sequences of M to one M plus a sequence of L operators,
            # same for m and l.
            if item == 'M' and item == op:
                op = 'L'
            elif item == 'm' and item == op:
                op = 'l'
            else:
                op = item
            if ops[op] == 0:  # Z, z
                result.extend([op, []])
        else:
            result.extend(split_floats(op, ops[op], item))
            op = result[-2]  # Remember last op

    return result


def convert_quadratic_to_cubic_path(q0, q1, q2):
    """
    Convert a quadratic Bezier curve through q0, q1, q2 to a cubic one.
    """
    c0 = q0
    c1 = (q0[0] + 2. / 3 * (q1[0] - q0[0]), q0[1] + 2. / 3 * (q1[1] - q0[1]))
    c2 = (c1[0] + 1. / 3 * (q2[0] - q0[0]), c1[1] + 1. / 3 * (q2[1] - q0[1]))
    c3 = q2
    return c0, c1, c2, c3


# ***********************************************
# Helper functions for elliptical arc conversion.
# ***********************************************

def vector_angle(u, v):
    d = hypot(*u) * hypot(*v)
    c = (u[0] * v[0] + u[1] * v[1]) / d
    if c < -1:
        c = -1
    elif c > 1:
        c = 1
    s = u[0] * v[1] - u[1] * v[0]
    return degrees(copysign(acos(c), s))


def end_point_to_center_parameters(x1, y1, x2, y2, fA, fS, rx, ry, phi=0):
    '''
    See http://www.w3.org/TR/SVG/implnote.html#ArcImplementationNotes F.6.5
    note that we reduce phi to zero outside this routine
    '''
    rx = fabs(rx)
    ry = fabs(ry)

    # step 1
    if phi:
        phi_rad = radians(phi)
        sin_phi = sin(phi_rad)
        cos_phi = cos(phi_rad)
        tx = 0.5 * (x1 - x2)
        ty = 0.5 * (y1 - y2)
        x1d = cos_phi * tx - sin_phi * ty
        y1d = sin_phi * tx + cos_phi * ty
    else:
        x1d = 0.5 * (x1 - x2)
        y1d = 0.5 * (y1 - y2)

    # step 2
    # we need to calculate
    # (rx*rx*ry*ry-rx*rx*y1d*y1d-ry*ry*x1d*x1d)
    # -----------------------------------------
    #     (rx*rx*y1d*y1d+ry*ry*x1d*x1d)
    #
    # that is equivalent to
    #
    #          rx*rx*ry*ry
    # = -----------------------------  -    1
    #   (rx*rx*y1d*y1d+ry*ry*x1d*x1d)
    #
    #              1
    # = -------------------------------- - 1
    #   x1d*x1d/(rx*rx) + y1d*y1d/(ry*ry)
    #
    # = 1/r - 1
    #
    # it turns out r is what they recommend checking
    # for the negative radicand case
    r = x1d * x1d / (rx * rx) + y1d * y1d / (ry * ry)
    if r > 1:
        rr = sqrt(r)
        rx *= rr
        ry *= rr
        r = x1d * x1d / (rx * rx) + y1d * y1d / (ry * ry)
    r = 1 / r - 1
    if -1e-10 < r < 0:
        r = 0
    r = sqrt(r)
    if fA == fS:
        r = -r
    cxd = (r * rx * y1d) / ry
    cyd = -(r * ry * x1d) / rx

    # step 3
    if phi:
        cx = cos_phi * cxd - sin_phi * cyd + 0.5 * (x1 + x2)
        cy = sin_phi * cxd + cos_phi * cyd + 0.5 * (y1 + y2)
    else:
        cx = cxd + 0.5 * (x1 + x2)
        cy = cyd + 0.5 * (y1 + y2)

    # step 4
    theta1 = vector_angle((1, 0), ((x1d - cxd) / rx, (y1d - cyd) / ry))
    dtheta = vector_angle(
        ((x1d - cxd) / rx, (y1d - cyd) / ry),
        ((-x1d - cxd) / rx, (-y1d - cyd) / ry)
    ) % 360
    if fS == 0 and dtheta > 0:
        dtheta -= 360
    elif fS == 1 and dtheta < 0:
        dtheta += 360
    return cx, cy, rx, ry, -theta1, -dtheta


def bezier_arc_from_centre(cx, cy, rx, ry, start_ang=0, extent=90):
    if abs(extent) <= 90:
        nfrag = 1
        frag_angle = float(extent)
    else:
        nfrag = int(ceil(abs(extent) / 90.))
        frag_angle = float(extent) / nfrag

    frag_rad = radians(frag_angle)
    half_rad = frag_rad * 0.5
    kappa = abs(4. / 3. * (1. - cos(half_rad)) / sin(half_rad))

    if frag_angle < 0:
        kappa = -kappa

    point_list = []
    theta1 = radians(start_ang)
    start_rad = theta1 + frag_rad

    c1 = cos(theta1)
    s1 = sin(theta1)
    for i in xrange(nfrag):
        c0 = c1
        s0 = s1
        theta1 = start_rad + i * frag_rad
        c1 = cos(theta1)
        s1 = sin(theta1)
        point_list.append((cx + rx * c0,
                          cy - ry * s0,
                          cx + rx * (c0 - kappa * s0),
                          cy - ry * (s0 + kappa * c0),
                          cx + rx * (c1 + kappa * s1),
                          cy - ry * (s1 - kappa * c1),
                          cx + rx * c1,
                          cy - ry * s1))
    return point_list


def bezier_arc_from_end_points(x1, y1, rx, ry, phi, fA, fS, x2, y2):
    if phi:
        # Our box bezier arcs can't handle rotations directly
        # move to a well known point, eliminate phi and transform the other point
        mx = mmult(rotate(-phi), translate(-x1, -y1))
        tx2, ty2 = transformPoint(mx, (x2, y2))
        # Convert to box form in unrotated coords
        cx, cy, rx, ry, start_ang, extent = end_point_to_center_parameters(
            0, 0, tx2, ty2, fA, fS, rx, ry
        )
        bp = bezier_arc_from_centre(cx, cy, rx, ry, start_ang, extent)
        # Re-rotate by the desired angle and add back the translation
        mx = mmult(translate(x1, y1), rotate(phi))
        res = []
        for x1, y1, x2, y2, x3, y3, x4, y4 in bp:
            res.append(
                transformPoint(mx, (x1, y1)) + transformPoint(mx, (x2, y2)) +
                transformPoint(mx, (x3, y3)) + transformPoint(mx, (x4, y4))
            )
        return res
    else:
        cx, cy, rx, ry, start_ang, extent = end_point_to_center_parameters(
            x1, y1, x2, y2, fA, fS, rx, ry
        )
        return bezier_arc_from_centre(cx, cy, rx, ry, start_ang, extent)



__code__='Gas1]gMYb8&=k*r=]Sct;;Cf01i1UId6jlm2\\\\\\X1f0]#!3I/:&dM)<K\'\'.O`Y?NG=11d^:Cba<r^_O6T>B.Y7@uV.Mm8]b&U"D[h3i=rL3rW4R/:0dq\'4/^0/@be4Fj&#!!9e;[`8Ss!i1?Vi5P>4O9Qg:\'@6es!&aio!*au3J8N(XJ4kD7!%nAF"mc5\\"7-)\\#@%<-&F]bZ7t;Z`IKK9ni+IQV!A=J&%X=CDcFu!#li(f=iE,$I^h\\I:EX\'qnfmmQG`61AQo=5%NM;8C6JlD=jLFY(DI9DHPE5"TD/s%TK<=r2r7hph_*NoWk-N>+Z/1&AQ(n.d=Me7V(J,rpCOFSPV(LEN&!KLi\\$plu95D\'aEN`fqm%6-*c]2uTODW\\2(O]7kt[UF0q^R"a9NuecQYMF]t\\7_*84&Q;eX)]AU/K1%7\'&Bld/!odL.?`+!k<#9?Q2Crg761=q;mDj-\'mt)G\\a?KTl/sP_caNZ`*Q<;RZB`GnA;@"DAR&HbE?Nu0.lsE/2OuBDJsPjQ_-$P_g%PAaMNaFZ9>:FH\'#Z.FW_IIJl^V*g_i6GhRa%#bb^6*^Z#2dh=fShJCtTcqPE?mIG$r?f.$Io1SLG`h[@M3]q<_pfIFmp0*#pr<g5LY9-VVR@9pe;WE7X`8GS93g:$bj4EgfFSYei4Yr%ATp;NO=5DK1)d.Q,=i:&]VA6"/CEn+Bm<Y.doo^/R+*VUO$qP,S#0lag;*R7hp^q@4oR\\5D$$UkoS[0;lkBhe>TA.-tcE_VH?$%L\'!G^#b_B<c??p]6`Z6h-8e-k*)NdMXiIK;C+V&$*??0\'OO:=4uoIu,huOPK%uqL$"i*2ST%YbHVq(IGEB8h[A$n"Zi1HdE^FsT%uQW2S`W1a4,@@&*:unOY)h!o^*c^7mRW\'G($S*_Bp?V;Mr\'8W035F1TL3>f2(RsGW8<>3!3@u^[3Smc#/(UcfW8%!L,3!(rPg88FGj;RMD&^?P[aL?<A\'bi-+QZBBN6L\'K^/\'Eck@-lY>m4MILsO&<?YS*>;dZWa*B--9!PR%Al[4XWMOEZJ"j+g3\':R1\'#j\\/U+e[>=:n9f?V<q$$i~>'
